import io
from lxml import etree

from ..Utils.filesystem import *
from .object import Object

class Level():
    def __init__(this, xml : str | bytes | File | etree.ElementBase, image : str | bytes | File , filesystem : Filesystem | Folder = None, gamepath : str = None, assets : str = '/assets') -> None:
        this.gamepath = gamepath
        this.assets = assets
        if this.assets == None:
            this.assets = '/assets'
        
        try:
            this.filesystem = filesystem
            if isinstance(this.filesystem, Filesystem):
                this.gamepath = this.filesystem.gamepath
                this.assets = this.filesystem.assets
                
            elif isinstance(this.filesystem, Folder):
                pass

            else:
                this.filesystem = Filesystem(this.gamepath, this.assets)
                this.filesystem.getAssets()
        except:
            raise FileNotFoundError('Must have a valid `filesystem` or `gamepath`')
        
        if isinstance(xml, bytes):
            this.xml_file = io.BytesIO(xml)
        elif isinstance(xml, File):
            this.xml_file = xml.rawcontent
        elif not hasattr(xml, 'read') and not isinstance(xml, str):
            raise TypeError(f"file can only 'str', 'bytes', or file-like object.")
        
        this.xml = etree.parse(this.xml_file).getroot()
        
        this.objects = []
        this.properties = {}
        this.room = (0,0)
        
        this.read()
    
    def read(this):
        this.objects = []
        this.properties = {}
        
        for element in this.xml:
            # comment safe-guard
            if element is etree.Comment:
                continue
            
            if element.tag == 'Object':
                properties = {}
                pos = (0,0)
                name = element.get('name')
                
                for el in element:
                    if not el is etree.Comment:
                        if el.tag == 'AbsoluteLocation':
                            pos = el.get('value')
                        
                        elif el.tag == 'Properties':
                            for property in el:
                                if not property is etree.Comment and property.tag == 'Property':
                                    properties[property.get('name')] = property.get('value')
                
                obj = Object(
                    this.filesystem.get(properties['Filename']), # get file because `Object` does not take filepath
                    filesystem = this.filesystem,
                    properties = properties,
                    pos = pos,
                    name = name,
                )
                
                this.objects.append(obj)
            
            if element.tag == 'Properties':
                for el in element:
                    if el is etree.Comment:
                        continue
                    
                    if el.tag == 'Property':
                        this.properties[el.get('name')] = el.get('name')
            
            if element.tag == 'Room':
                for el in element:
                    if el is etree.Comment:
                        continue
                    
                    if el.tag == 'AbsoluteLocation':
                        this.room = tuple([float(_) for _ in el.get('value').split(' ')])
    
    def getObject(this, name : str):
        """
        Get object by name

        Args:
            name (str): Object name.
        """
        
        for obj in this.objects:
            if obj.name == name:
                return obj
        
        return None



from lxml import etree

def findTag(root : etree.ElementBase, tag : str) -> etree.Element:
    element = 0
    curTag = ''
    for e in root:
        print(e.tag)
        if e.tag is etree.Comment:
            continue
        curTag = e.tag
        if curTag == tag:
            return e
        
    return None



__version__ = "0.0.1-alpha"
__author__ = 'ego-lay-atman-bay'

from .Game import Game
from . import classes
from . import Font
from .classes import widget
from . import Utils

def load(gamepath : str, assets : str = '/assets', db : str = '/Data/water.db', profile : str = None):
    """load game

    Args:
        gamepath (str): Folder to extracted game.
        assets (str, optional): Relative path to assets folder. Defaults to '/assets'.
        db (str, optional): Relative path to database file from assets folder. Defaults to '/Data/water.db'.
        profile (str, optional): Relative path to profile file in WMW2. Defaults to `None`
    """
    
    return Game(gamepath=gamepath, assets=assets, db=db, profile=profile)



import lxml
from lxml import etree
import io

from ..gameobject import GameObject
from ..Utils import Texture
from .sprite import Sprite
from ..Utils import XMLTools
from ..Utils.filesystem import *

class Object(GameObject):
    def __init__(
        this,
        file : str | bytes | File ,
        filesystem : Filesystem | Folder = None,
        gamepath : str = None,
        assets : str = '/assets',
        properties : dict = {},
        pos : tuple | str = (0,0),
        name : str = 'Obj'
    ) -> None:
        """Get game object. Game object is `.hs` file.

        Args:
            gamepath (str): Game path
            assets (str): Assets path, relative to game path
            object (str): Object file relative to assets path. Must be `.hs` file.
            properties (dict, optional): Object properties that override default properties. Defaults to {}.
            position ((tuple, str), optional): Object position. Can be string or tuple. Defaults to (0,0).
            name (str): The name of the object. Defaults to `'Obj'`
        """
        
        super().__init__(filesystem, gamepath, assets)
        
        this.file = super().test_file(file)
        
        this._properties = properties
        if isinstance(pos, str):
            this.pos = tuple([float(a) for a in pos.split(' ')])
        else:
            this.pos = tuple(pos)
        
        this.xml : etree.ElementBase = etree.parse(this.file).getroot()
        this.sprites = []
        this.shapes = []
        this.UVs = []
        this.VertIndices = []
        this.defaultProperties = {}
        this.properties = {}
        this.type = None
        this.name = name
        
        this.image = None
        
        this.readXML()
        
    def readXML(this):
        # specifically specifying type so it's easier to use in vscode
        
        this.getSprites()
        this.getProperties()
        this.getShapes()
        this.getUVs()
        this.getVertIndices()
        
    def getSprites(this):
        spritesXML = XMLTools.findTag(this.xml, 'Sprites')
        if spritesXML == None:
            return None
        for sprite in spritesXML:
            if (not sprite is etree.Comment) and sprite.tag == 'Sprite':
                attributes = sprite.attrib
                this.sprites.append(Sprite(
                    file = this.filesystem.get(attributes['filename']),
                    filesystem = this.filesystem,
                    attributes = attributes
                ))
        
        return this.sprites
    
    def getProperties(this):
        this.getDefaultProperties()
        for prop in this.defaultProperties:
            this.properties[prop] = this.defaultProperties[prop]
        for prop in this._properties:
            this.properties[prop] = this._properties[prop]
        return this.properties
    
    def getDefaultProperties(this):
        this.defaultProperties = {}
        propertiesXML = XMLTools.findTag(this.xml, 'DefaultProperties')
        if propertiesXML == None:
            return None
        for prop in propertiesXML:
            if (not prop is etree.Comment) and prop.tag == 'Property':
                name = prop.get('name')
                value = prop.get('value')
                if name == 'Type':
                    this.type = value
                this.defaultProperties[name] = value
            return this.defaultProperties
    
    def getShapes(this):
        this.shapes = []
        shapesXML = XMLTools.findTag(this.xml, 'Shapes')
        if shapesXML == None:
            return None
        for shape in shapesXML:
            if (not shape is etree.Comment) and shape.tag == 'Shape':
                obj = Shape()
                obj.readXML(shape)
                this.shapes.append(obj)
        return this.shapes
    
    def getUVs(this):
        this.UVs = []
        UVsXML = XMLTools.findTag(this.xml, 'UVs')
        if UVsXML == None:
            return None
        for UV in UVsXML:
            if (not UV is etree.Comment) and UV.tag == 'UV':
                pos = UV.get('pos')
                this.UVs.append(tuple([float(axis) for axis in pos.split(' ')]))
        return this.UVs
    
    def getVertIndices(this):
        pass
        

class Shape():
    def __init__(self) -> None:
        pass
    
    def readXML(this, xml : etree.ElementBase = None):
        pass
    
    class Point():
        def __init__(self) -> None:
            pass        



from .Widget import Widget

from .WT_PUSH_BUTTON import WT_PUSH_BUTTON
from .WT_LABEL import WT_LABEL
from .WT_TOGGLE import WT_TOGGLE
from .WT_SLIDER import WT_SLIDER
from .WT_PROGRESS_BAR import WT_PROGRESS_BAR
from .WT_GROUP import WT_GROUP
from .WT_ICON_LIST import WT_ICON_LIST
from .WT_SCROLLABLE_CAMERA import WT_SCROLLABLE_CAMERA
from .WT_SCROLLABLE_SET import WT_SCROLLABLE_SET
from .WT_CANVAS import WT_CANVAS
from .WT_FINGER_CATCHER import WT_FINGER_CATCHER

WidgetTypes = {
    'WT_PUSH_BUTTON': WT_PUSH_BUTTON,
    'WT_LABEL': WT_LABEL,
    'WT_TOGGLE': WT_TOGGLE,
    'WT_SLIDER': WT_SLIDER,
    'WT_PROGRESS_BAR': WT_PROGRESS_BAR,
    'WT_GROUP': WT_GROUP,
    'WT_ICON_LIST': WT_ICON_LIST,
    'WT_SCROLLABLE_CAMERA': WT_SCROLLABLE_CAMERA,
    'WT_SCROLLABLE_SET': WT_SCROLLABLE_SET,
    'WT_CANVAS': WT_CANVAS,
    'WT_FINGER_CATCHER': WT_FINGER_CATCHER,
}



from ..Utils.textures import getHDFile, getTextueSettings, getTexture
from ..Utils.filesystem import *
from ..Utils import joinPath, Texture
from ..gameobject import GameObject

import numpy
import PIL.Image
from lxml import etree

import io
import os

class Imagelist(GameObject):
    class Type():
        IMAGELIST = 0
        PAGES = 1
    
    def __init__(
        this,
        file : str | bytes | File,
        filesystem : Filesystem | Folder = None,
        gamepath : str = None, assets : str = '/assets',
        HD : bool = False
    ) -> None:
        """
        Get imagelist from file
        Args:
            file (str | bytes | File): File to read.
            filesystem (Filesystem | Folder, optional): Filesystem to use. Defaults to None.
            HD (bool, optional): Whether to use HD textures. Defaults to False.
            gamepath (str, optional): Path to game. Only is used when filesystem is `Folder` or `None`. Defaults to None.
            assets (str, optional): Path to assets relative to gamepath. Defaults to '/assets'.
        Raises:
            FileNotFoundError: Filesystem is not usable and no gamepath.
        """
        
        super().__init__(filesystem, gamepath, assets)
        
        this.file = super().test_file(file)

        this.HD = HD
        this.xml = etree.parse(this.file).getroot()
        this.textureBasePath = '/Textures/'
        this.pages = []
        this.type = this.Type.IMAGELIST

        this.images = {}

        this.read()
    
    def read(this):
        this.type = this.Type.IMAGELIST
        
        for element in this.xml:
            if element is etree.Comment:
                continue
            
            if element.tag == 'Page':
                this.type = this.Type.PAGES
                page = this.Page(
                    element,
                    filesystem = this.filesystem,
                    HD = this.HD,
                )
                
                this.pages.append(page)
        
        if this.type == this.Type.IMAGELIST:
            page = this.Page(
                this.xml,
                filesystem = this.filesystem,
                HD = this.HD,
            )
            
            this.pages.append(page)
    
    def getImage(this, name : str):
        for page in this.pages:
            image = page.getImage(name)
            if image:
                return image
            
        return None
        
        # return this.filesystem.get(os.path.join(this.textureBasePath, name))
    
    class Page(GameObject):
        def __init__(
            this,
            element : etree.ElementBase,
            filesystem: Filesystem | Folder = None,
            gamepath: str = None,
            assets: str = '/assets',
            HD = False,
        ) -> None:
            super().__init__(filesystem, gamepath, assets)
            
            this.xml = element
            
            this.HD = HD
            this.id = ''
            this.imgSize = (1,1)
            this.textureBasePath = '/Textures/'
            this.file = ''
            this.atlas = None
            this.images = {}
            
            this.read()
        
        class Image():
            def __init__(
                this,
                atlas : PIL.Image.Image,
                properties : dict
            ) -> None:
                """Image for Imagelist

                Args:
                    this (_type_): _description_
                    atlas (PIL.Image.Image): Atlas file containing all images
                    properties (dict): Properties for Image
                """

                this.atlas = atlas
                this.properties = properties

                this.size = (1,1)
                this.offset = (0,0)
                this.rect = (0,0,0,0)
                this.name = ''

                this.image = PIL.Image.new('RGBA', this.size)

                this.rawdata = io.BytesIO()

                this.getData()
                this.getImage()

            def getData(this):
                if 'size' in this.properties:
                    this.size = tuple([int(v) for v in this.properties['size'].split(' ')])
                if 'offset' in this.properties:
                    this.offset = tuple([int(v) for v in this.properties['offset'].split(' ')])
                if 'rect' in this.properties:
                    this.rect = tuple([int(v) for v in this.properties['rect'].split(' ')])
                if 'name' in this.properties:
                    this.name = this.properties['name']

            def getImage(this):
                this.image = this.atlas.crop(numpy.add(this.rect, (0,0) + this.rect[0:2]))
                this.image = this.image.resize(this.size)
                
                this.image.save(this.rawdata, format = os.path.splitext(this.name)[1][1::])
                return this.image

            def show(this):
                this.image.show()


        def read(this):
            this.attributes = this.xml.attrib
            if 'imgSize' in this.attributes:
                this.size = tuple([int(v) for v in this.attributes['imgSize'].split(' ')])
            if 'textureBasePath' in this.attributes:
                this.textureBasePath = this.attributes['textureBasePath']
            if 'file' in this.attributes:
                this.file = this.attributes['file']
            if 'id' in this.attributes:
                this.id = this.attributes['id']
    
            if this.HD:
                hd = getHDFile(this.file)
                if this.filesystem.exists(hd):
                    this.file = hd
    
            this.fullAtlasPath = ''
    
            if this.gamepath:
                this.fullAtlasPath = joinPath(this.gamepath, this.assets, this.file)
                print(this.fullAtlasPath)
    
            this.getAtlas()
            this.getImages()
    
    
        def getAtlas(this):
            if this.filesystem.exists(this.file):
                file = this.filesystem.get(this.file)
                image = Texture(file.read())
                this.atlas = image.image.copy()
            else:
                this.textureSettings = getTextueSettings(
                    this.gamepath,
                    this.assets,
                    joinPath(os.path.dirname(os.path.dirname(this.textureBasePath)), 'Data/textureSettings.xml'),
                    this.name
                )
    
                this.atlas = getTexture(this.fullAtlasPath, this.textureSettings, this.size)
    
        def getImages(this):
            for image in this.xml:
                if not image.tag is etree.Comment:
                    if image.tag == 'Image':
                        texture = this.Image(this.atlas, properties = image.attrib)
                        this.images[image.get('name')] = texture
                        
                        # print(f'{this.textureBasePath = }')
                        # print(f'{texture.name = }')
    
                        this.filesystem.add(
                            joinPath(this.textureBasePath, texture.name),
                            texture.rawdata.getvalue(),
                            replace = True
                        )
    
        def getImage(this, name : str) -> Image:
            if name in this.images:
                return this.images[name]
            else:
                return None
    
        def getNO_TEX(this):
            NO_TEX_settings = getTextueSettings(
                this.gamepath,
                this.assets,
                joinPath(os.path.dirname(os.path.dirname(this.textureBasePath)), 'Data/textureSettings.xml'),
                os.path.join(this.textureBasePath, 'NO_TEX.png')
            )
            NO_TEX_image = PIL.Image.open(joinPath(this.gamepath, this.assets, this.textureBasePath, 'NO_TEX.png')).convert('RGBA')
            this.NO_TEX = this.Image(NO_TEX_image, {
                'size': ' '.join([str(x) for x in NO_TEX_image.size]),
                'rect': ' '.join([str(x) for x in (0,0) + NO_TEX_image.size]),
                'name': 'NO_TEX.png',
            })
            # this.Image(this.atlas, image.attrib)
    
    




import pathlib
import os
import io
from filetype import filetype
from PIL import Image
import zipfile

from .path import joinPath
# from . import Waltex
# from . import ImageUtils
# from ..classes import sprite
# from ..classes import Object

__all__ = ['Filesystem', 'File', 'Folder', 'FileBase']


FILE_READERS = []

# Filesystem object.

class Filesystem():
    def __init__(this, gamepath : str, assets : str) -> None:
        this.gamepath = gamepath
        this.assets = assets
        this.parent = this
        this.root = Folder(this, '/')
    
    def get(this, path : str):
        return this.root.get(path)
    
    def add(this, path : str, file : str | bytes, replace = False):
        """
        Adds file with path to Folder.

        Args:
            path (str): Path to destenation of file. Must be str.
            file (str | bytes): File. This can be a path to the file to use, raw `bytes` of the file, or file-like object.
            replace (bool, optional): Replace any conflicting files? Defaults to False.

        Raises:
            TypeError: If file is not `str`, `bytes` or file-like object.

        Returns:
            File: File object of the added file.
        """
        if isinstance(file, str):
            with open(file, 'rb') as f:
                file = f.read()
            # print('file')
        elif isinstance(file, bytes):
            # print('bytes')
            pass
        elif hasattr(file, 'read'):
            file = file.read()
            if isinstance(file, str):
                file = file.encode()
            # print('file-like')
        else:
            raise TypeError(f"file can only 'str', 'bytes', or file-like object.")
        
        return this.root.add(path = path, content = file, replace = replace)
        
    def getAssets(this, extract_zip = False, split_imagelist = False):
        """
        Scans the assets folder and adds all the files into the filesystem. Prepare for hundreds of files being opened.
        """
        
        print(this.gamepath)
        print(f'{this.gamepath = }\n{this.assets = }')
        assets = joinPath(this.gamepath, this.assets)
        
        for dir, subdir, files in os.walk(assets):
            for file in files:
                path = pathlib.Path('/', os.path.relpath(os.path.join(dir, file), assets)).as_posix()
                print(path)
                # print(os.path.join(dir, file))
                fileobj : File = this.add(path, os.path.join(dir, file))
                
                if fileobj.extension == 'zip' and extract_zip:
                    fileobj.read()
        
        return this
    
    def exists(this, fp : str):
        return this.root.exists(fp)
    
# Filesystem helpers
class FileBase():
    name = ''
    
    def __init__(this, parent, path : str):
        """File Base

        Args:
            parent (Folder): Parent. Use `None` for root.
            path (str): File path.
        """
        this._type = this._Type(None)
        this.name = pathlib.Path(path).parts[0]
        this.parent = parent
        
    @property
    def path(this):
        if this.parent == None or isinstance(this.parent, Filesystem):
            return '/'
        return pathlib.Path(this.parent.path, this.name).as_posix()
    
    @property
    def root(this):
        if this.parent == None or this.name == '':
            return this
        return this.parent.root
    
    @property
    def filesystem(this):
        if isinstance(this.parent, Filesystem):
            return this.parent
        if this.parent == None or this.name == '':
            return this
        return this.parent.filesystem
        
    class _Type():
        FOLDER = 0
        FILE = 1
        
        def __init__(this, type : int) -> None:
            this.value = type
    
class File(FileBase):
    def __init__(this, parent, path: str, content : bytes):
        """File

        Args:
            parent (Folder): Parent. Use `None` for root.
            path (str): File path.
            content (bytes): Contents of file as bytes.
        """
        super().__init__(parent, path)
        this._type.value = this._Type.FILE
        this._rawcontent = content
        
        this.rawcontent = io.BytesIO(content)
        # seek back to the start to be able to read the data later.
        this.rawcontent.seek(0)
        
        this.content = None
        
        this.testFile()
        
    def testFile(this):
        """Tests what type of file this is."""
        this.type = filetype.guess(this.rawcontent.getvalue())
        
        if this.type == None:
            this.type = None
            this.extension = os.path.splitext(this.name)[1][1::]
            if not this.extension:
                this.mime = f'text/raw'
            else:
                this.mime = f'text/{this.extension}'
            
        else:
            this.mime = this.type.mime
            this.extension = this.type.extension
            
        return this.mime
        
    def read(this, **kwargs):
        this.rawcontent.seek(0)
        
        reader = Reader()
        
        for r in FILE_READERS:
            print(r)
            if r.check(this.mime, this.extension, this.rawcontent, filesystem = this.filesystem, **kwargs):
                reader = r
                break
        
        this.content = reader.read(this.mime, this.extension, this.rawcontent, **kwargs)
        
        # if this.mime == 'image/waltex':
        #     this.content = Waltex(this.rawcontent.getvalue())
        #     this.image = this.content.image
        
        # elif this.mime.startswith('image/'):
        #     this.content = Image.open(this.rawcontent.getvalue())
        #     this.image = this.content
        
        # elif this.extension == 'zip':
        #     this.content = zipfile.ZipFile(this.rawcontent)
        #     if 'extract' in kwargs and kwargs['extract']:
        #         print(f'extracting {this.name}')
                
        #         files = this.content.namelist()
        #         for f in files:
        #             print(f)
                    
        #             content = this.content.read(f)
        #             this.root.add(f, content, replace = True)
                
        # elif this.mime.startswith('text/'):
        #     if this.extension == 'imagelist':
        #         this.content = ImageUtils.Imagelist(this.rawcontent.getvalue(), this.root, **kwargs)
        #         # I need to make Imagelist() accept a Folder or Filesystem, and raw file data.
        #         # raise NotImplementedError('Imagelist reading is currently not implemented yet.')
        #         # this.content = ImageUtils.Imagelist()
        #     else:
        #         this.content = this.rawcontent.read().decode(encoding=kwargs['encoding'])
        
        this.rawcontent.seek(0)
        
        return this.content
    
    def get(this, path : str):
        return this.parent.get(path)
    
    def add(this, path : str, file : bytes, replace = False):
        return this.parent.add(path, file, replace = replace)
    
    def exists(this, path : str):
        return this.parent.exists(path)

class Folder(FileBase):
    def __init__(this, parent = None, path: str = None):
        """Folder

        Args:
            this (_type_): _description_
            parent (Folder): Parent. Use `None` for root.
            path (str): Folder path.
        """
        if isinstance(parent, str) and path == None:
            path = parent
            parent = None
            
        if not path:
            path = '/'
        
        super().__init__(parent, path)
        this._type.value = this._Type.FOLDER
        this.files = []
        
    def add(this, path : str, content : bytes, replace = False) -> File:
        parts = pathlib.Path(path).parts
        
        file = this._getPath(pathlib.Path(*parts).as_posix())
        if len(parts) > 1:
            if file == None:
                file = Folder(this, parts[0])
                this.files.append(file)
                
            if file._type.value != file._Type.FOLDER:
                raise NotADirectoryError(f"{file.path} is not a directory.")
            
            return file.add(pathlib.Path(*parts[1::]).as_posix(), content, replace=replace)
        else:
            if file != None:
                if  not replace:
                    raise FileExistsError(f'File {file.path} already exists.')
                print(f'File {file.path} already exists. Now replacing it.')
                this.files.remove(file)
            
            file = File(this, parts[0], content = content)
            this.files.append(file)
            
            return file
            
        
    def _getPath(this, path : str):
        parts = pathlib.Path(path).parts
        file = None
        if parts[0] == '\\':
            file = this.root
        elif parts[0] == '..':
            file = this.parent
        else:
            for f in this.files:
                if f.name == parts[0]:
                    file = f
                    break
        return file
    
    def get(this, path : str):
        parts = pathlib.Path(path).parts
        if len(parts) == 0:
            return this
        
        file = this._getPath(path)
        if file == None:
            return file
        if file._type.value == this._Type.FOLDER:
            file = file.get(pathlib.Path(*parts[1::]))
        return file
    
    def exists(this, path : str):
        return this.get(path) != None

class Reader():
    MIME = 'text/'
    EXTENSION = 'txt'
    
    def __init__(self, ) -> None:
        pass
    
    def check(this, mime : str, extension : str, rawdata : io.BytesIO, **kwargs):
        return mime.startswith(this.MIME)
    
    def read(this, mime : str, extension : str, rawdata : io.BytesIO, **kwargs):
        if 'endoding' in kwargs:
            encoding = kwargs['encoding']
        else:
            encoding = 'utf-8'
        
        return rawdata.getvalue().decode(encoding=encoding)

def register_reader(reader : Reader):
    """
    Add reader for file type.
    reader must inherit from `Utils.filesystem.Reader` class.

    Args:
        reader (Reader): Inherited `Reader` class.
    """
    
    if isinstance(reader, Reader):
        FILE_READERS.append(reader)
    else:
        raise TypeError('reader must inherit from `Utils.filesystem.Reader`')



import os
import pathlib
import lxml
from lxml import etree
import natsort
import zipfile

from .Utils import Filesystem
from .Utils import Texture
from .Utils.path import joinPath
from .classes import *

os.path.isdir

class Game():
    def __init__(this, gamepath : str, assets : str = '/assets', db : str = '/Data/water.db', profile : str = None) -> None:
        """load game

        Args:
            gamepath (str): Folder to extracted game.
            assets (str, optional): Relative path to assets folder. Defaults to '/assets'.
            db (str, optional): Relative path to database file from assets folder. Defaults to '/Data/water.db'.
            profile (str, optional): Relative path to profile file in WMW2. Defaults to `None`
        """
        this.gamepath = os.path.abspath(gamepath)
        print(f'{gamepath = }\n{this.gamepath = }')
        this.assets = assets
        this.db = db
        this.profile = profile
        
        this.updateFilesystem()
        
    def updateFilesystem(this):
        this.filesystem = Filesystem(this.gamepath, this.assets)
        this.filesystem.getAssets()
        
    def loadLevel(this, xmlPath : str = None, imagePath : str = None, ):
        """
        Load level

        Args:
            xmlPath (str, optional): Path to xml file. Defaults to None.
            imagePath (str, optional): Path to image file. Defaults to None.
        """
        xml = None
        if xmlPath:
            xml = this.filesystem.get(xmlPath)
        
        image = None
        if imagePath:
            image = this.filesystem.get(imagePath)
        
        return Level(
            xml=xml,
            image=image,
            filesystem=this.filesystem,
        )
    
    def loadObject(
        this,
        object : str,
        **kwargs
    ):
        """
        Loads object

        Args:
            object (str): Path to `.hs` object file.

        Returns:
            classes.object.Object: Where's My Water? object.
        """
        
        return Object(
            this.filesystem.get(object),
            filesystem = this.filesystem,
            **kwargs
        )
    
    def loadImagelist(
        this,
        imagelist : str,
        HD = False,
    ):
        """
        Load imagelist

        Args:
            imagelist (str): Path to `.imagelist` file.
            HD (bool, optional): Whether to use HD textures. Defaults to False.

        Returns:
            classes.imagelist.Imagelist: Imagelist object.
        """
        return Imagelist(
            this.filesystem.get(imagelist),
            filesystem = this.filesystem,
            HD = False,
        )
    
    def loadSprite(
        this,
        sprite : str,
        **kwargs
    ):
        """
        Loads sprite.

        Args:
            sprite (str): Path to `.sprite` file.`

        Returns:
            classes.sprite.Sprite: Sprite object.
        """
        return Sprite(
            this.filesystem.get(sprite),
            filesystem = this.filesystem,
            **kwargs
        )
    
    def loadTexture(
        this,
        texture : str,
    ):
        """
        Get image texture. Doesn't matter if it's a `.waltex` image or not.

        Args:
            this (_type_): _description_
            texture (str): Path to image file.

        Returns:
            Utils.textures.Texture: Texture object.
        """
        return Texture(
            this.filesystem.get(texture)
        )
    
    def loadLayout(this, layout : str):
        raise NotImplementedError('load layout is not implemented yet.')
    
    def generateFileManifest(this, writeFile : bool = True):
        manifest = []
        assets = joinPath(this.gamepath, this.assets)
        for dir, subdir, files in os.walk(assets):
            for file in files:
                # print(f'{file = }\n{dir = }\n{subdir = }')
                
                path = pathlib.Path('/', os.path.relpath(os.path.join(dir, file), assets)).as_posix()
                # path = pathlib.Path(path).parts
                # print(f'{path = }')
                manifest.append(path)
                
        manifest = natsort.natsorted(manifest)
        
        content = '\n'.join(manifest)
        
        path = joinPath(this.gamepath, this.assets, 'FileManifest.txt')
        # print(path)
        
        if writeFile:
            with open(path, 'w') as file:
                file.write(content)
                
        return manifest



from lxml import etree
from PIL import Image
import numpy

from .Imagelist import Imagelist
from ..Utils.filesystem import *
from ..gameobject import GameObject

class Sprite(GameObject):
    def __init__(this, file : str | bytes | File, filesystem: Filesystem | Folder = None, gamepath: str = None, assets: str = '/assets', attributes : dict = {}) -> None:
        super().__init__(filesystem, gamepath, assets)
        this.file = super().test_file(file)
        
        this.xml : etree.ElementBase = etree.parse(this.file).getroot()
        
        this.attributes = attributes
        this.animations = []
        
        this.readXML()
        
    def readXML(this):
        this.animations = []
        for element in this.xml:
            if (not element is etree.Comment) or element.tag == 'Animation':
                animation = this.Animation(
                    element,
                    this.filesystem
                )
                
                this.animations.append(animation)
        

    class Animation(GameObject):
        def __init__(this, xml : str | etree.ElementBase, filesystem: Filesystem | Folder = None, gamepath: str = None, assets: str = '/assets') -> None:
            super().__init__(filesystem, gamepath, assets)
            
            if isinstance(xml, str):
                this.xml : etree.ElementBase = etree.parse(xml).getroot()
            else:
                this.xml = xml
            
            this.properties = {}
            this.name = ''
            this.textureBasePath = '/Textures/'
            this.atlas : Imagelist = None
            this.fps = 30
            this.playbackMode = 'ONCE'
            this.loopCount = 1
            
            this.frames = []
            
            this.readXML()
            
        def readXML(this):
            this.getAttributes()
            this.getFrames()
        
        def getAttributes(this):
            this.properties = this.xml.attrib
            
            if 'name' in this.properties:
                this.name = this.properties['name']
            if 'textureBasePath' in this.properties:
                this.textureBasePath = this.properties['textureBasePath']
            if 'atlas' in this.properties:
                this.atlas = Imagelist(
                    this.filesystem.get(this.properties['atlas']),
                    this.filesystem,
                    HD=False
                )
                
                # this.atlasHD = Imagelist(
                #     this.filesystem.get(this.properties['atlas']),
                #     this.filesystem,
                #     HD=True
                # )
                
            if 'fps' in this.properties:
                this.fps = float(this.properties['fps'])
            if 'playbackMode' in this.properties:
                this.playbackMode = this.properties['playbackMode']
            if 'loopCount' in this.properties:
                this.loopCount = int(this.properties['loopCount'])
            
            
        def getFrames(this):
            this.frames = []
            
            if this.xml == None:
                return None
            for f in this.xml:
                if (not f is etree.Comment) and f.tag == 'Frame':
                    this.frames.append(this.Frame(
                        f.attrib,
                        this.atlas,
                        this.textureBasePath
                    ))
            
            return this.frames
        
        
        # Frame
        class Frame():
            def __init__(this, properties : dict, atlas : Imagelist = None, textureBasePath : str = None) -> None:
                this.atlas = atlas
                this.textueBasePath = textureBasePath
                this.properties = properties
                
                this.name = ''
                this.offset = (0,0)
                this.scale = (1,1)
                this.angleDeg = 0
                this.repeat = 0
                
                this.image = None
                
                this.getData()
                this.getImage()
            
            def getData(this):
                if 'name' in this.properties:
                    this.name = this.properties['name']
                if 'offset' in this.properties:
                    this.offset = tuple([float(x) for x in this.properties['offset'].split(' ')])
                if 'scale' in this.properties:
                    this.scale = tuple([float(x) for x in this.properties['scale'].split(' ')])
                if 'angleDeg' in this.properties:
                    this.angleDeg = float(this.properties['angleDeg'])
                if 'repeat' in this.properties:
                    this.repeat = int(this.properties['repeat'])
                
            def getImage(this):
                this._image = this.atlas.getImage(this.name)
                this.image = this._image.image.copy()
                
            def applyEffects(this):
                this.image = this._image.image.copy()
                this.image = this.image.resize(tuple([round(_) for _ in (numpy.array(this._image.size) * numpy.array(this.scale))]))
                this.image = this.image.rotate(this.angleDeg)



import os
import pathlib

def joinPath(*args : str) -> str:
    if len(args) == 0:
        return ''
    else:
        return pathlib.Path(getRelPath(args[0]), joinPath(*args[1:])).as_posix()
    # for p in args:
    #     path = os.path.join(path, getRelPath(p))
        
    # return path

def getRelPath(path : str) -> str:
    parts = pathlib.Path(path).parts
    if parts[0] == '\\':
        return pathlib.Path(*parts[1::]).as_posix()
    return path


import lxml
from lxml import etree
from PIL import Image, ImageTk

# main widget class
class Widget():
    """
        Main widget
    """
    def __init__(this, widget : etree.Element, gamePath, texturePath : str, screenSize : tuple) -> None:
        this.widget = widget
        this.attributes = this.widget.attr
        this.name = this.type = this.attributes['type']
        this.gamePath = gamePath
        
        this.pos = (0, 0)
        this.size = (0, 0)
        this.id = 0
        this.layer = 0
        
        this.forceAspect = False
        this.visible = True
        
        this.getValues()
        
        this.image = Image.new('RGBA', (100,100))
        
    def getValues(this):
        if this.attributes['pos']:
            this.pos = [float(v) for v in tuple(this.attributes['pos'].split(' '))]
        
        if (this.attributes['id']):
            this.id = float(this.attributes['id'])
            
        if (this.attributes['layer']):
            this.layer = float(this.attributes['layer'])
            
        if (this.attributes['size']):
            this.size = [float(v) for v in tuple(this.attributes['size'].split(' '))]
            
        if (this.attributes['forceAspect']):
            this.setForceAspect(this.attributes['forceAspect'])
            
        if (this.attributes['visible']):
            this.visible = bool(this.attributes['visible'])
        
    def setForceAspect(this, aspect = (1,1)):
        if isinstance(aspect, str):
            forceAspect = tuple([float(v) for v in aspect.split(':')])
        elif not aspect:
            forceAspect = False
        else:
            forceAspect = tuple(aspect)
        
        this.forceAspect = forceAspect


from .Utils.filesystem import *

import io

class GameObject():
    def __init__(this, filesystem : Filesystem | Folder = None, gamepath : str = None, assets : str = '/assets') -> None:
        this.gamepath = gamepath
        this.assets = assets
        if this.assets == None:
            this.assets = '/assets'
        
        # try:
        this.filesystem = filesystem
        if isinstance(this.filesystem, Filesystem):
            this.gamepath = this.filesystem.gamepath
            this.assets = this.filesystem.assets
            
        elif isinstance(this.filesystem, Folder):
            pass

        else:
            this.filesystem = Filesystem(this.gamepath, this.assets)
            this.filesystem.getAssets()
        # except Exception as e:
        #     print(f'Error: {str(e)}')
        #     raise FileNotFoundError('Must have a valid `filesystem` or `gamepath`')
    
    def test_file(this, file) -> io.BytesIO | str:
        if isinstance(file, bytes):
            file = io.BytesIO(file)
        elif isinstance(file, File):
            file = file.rawcontent
        elif not hasattr(file, 'read') and not isinstance(file, str):
            raise TypeError(f"file can only 'str', 'bytes', or file-like object.")
    
        return file
    


import os
import pathlib
import numpy
from lxml import etree
from .waltex import WaltexImage, Waltex
from PIL import Image
import io

from .path import joinPath
from .XMLTools import findTag
from .filesystem import Filesystem, Folder, File

_cachedWaltextImages = {}

class Texture():
    def __init__(this, image : Image.Image | Waltex | File) -> None:
        this._image = image
        
        if isinstance(this._image, File):
            this._image = this._image.read()
        
        if isinstance(this._image, Waltex):
            this.image = this._image.image
        elif isinstance(this._image, Image.Image):
            this.image = this._image
        else:
            raise TypeError('image must be PIL.Image.Image or Waltex')
        
    @property
    def size(this):
        return this.image.size

def getHDFile(file):
    split = os.path.splitext(file)
    split = list(split)
    split.insert(1, '-HD')
    return ''.join(split)
            
def getTexture(path : str, textureSettings : dict, size : tuple, cache = True) -> Image.Image:
    type = os.path.splitext(path)[1][1:]
    image = None
    if type == 'waltex':
        if cache:
            try:
                image = _cachedWaltextImages[path].copy()
            except:
                pass
        if image == None:
            image = Waltex(
                path
            ).image
            if cache:
                _cachedWaltextImages[path] = image.copy()
    else:
        image = Image.open(path).convert('RGBA')
        
    return image
    
def getTextueSettings(gamepath : str, assets : str, textureSettings : str, name : str, ) -> dict:
    fullpath = joinPath(gamepath, assets, textureSettings)
    xml = etree.parse(fullpath).getroot()
    
    Texture = etree.ElementBase()
    for i in xml:
        if not i.tag is etree.Comment:
            if i.tag == 'Texture' and i.get('name') == name:
                Texture = i
                break
            
    attributes = {
        'colorspace': 'RGBA4444',
        'premultiplyAlpha': False,
        'dePremultiplyAlpha': False,
    }
    
    values = Texture.attrib
    
    for key in values:
        attributes[key] = values[key]
    
    return attributes
